#include <stdio.h>
#include <ctype.h>
#define VERSION  "FILL - Dave Rand, Version 1.3  Sun Feb 15 13:25:37 1987"

/*
    Dave Rand
    72 Longfellow St.
    Thousand Oaks, CA  91360
    12/09/86

    AZTEC C86 - 3.20e
    
    This program may be freely distributed, in source and/or object,
    provided no fee is charged for it.
    
    FILL is a utility designed to allow you to fill floppies with multiple
    files. It will size the destination media automatically, and copy files
    to it in an efficient manner. New destination media is requested when the
    current media is full. If any files are too large to fit on the
    destination media, a message detailing this will appear on the screen.

    An invocation of fill would be as follows:
    
    fill [-q] source-path [source-path] destination-path

    Wildcard specifications are supported, as are paths.
    
    -q specifies quiet mode. In this mode, destination files are overwritten
    automatically if they exist. Normally, you are prompted if a file exists.
    
*/

#define MAXF 1024	/* Maximum number of files in any directory */
#define SETDMA	26
#define SRCHFIR	0x4e
#define SRCHNXT	0x4f

int Cpcdos4;		/* Indicates if we are running on CDOS 4.1 */
long Fsize;		/* Current file size in bytes */
int Curdrive;		/* Current drive number */
int Quiet=0;		/* Default to noisy */

main(argc,argv)
int argc;
char **argv;
{
    int c;

    if (argc < 3 ) {
	fprintf(stderr,
	"%s\nUsage:\nfill [-q] source-path [source-path] dest-path\n",
	VERSION);
	exit(1);
    }

    Cpcdos4 = (strcmp(getenv("OS"),"CPCDOS 4.1") == 0);
    
    Curdrive = bdos(25,0); /* since PCDOS does not restore the current drive
    			      on program termination, we must do it.
			    */

    for (c = 1; c < argc-1; c++) {
	if (argv[c][0] == '-')
	    switch(tolower(argv[c][1])) {
		case 'q':
		    Quiet = 1;
		    break;
		
		default:
		    fprintf(stderr,"fill: Unknown option: %s\n",argv[c]);
		    break;
	    }
    }
	    
	
    for (c = 1; c < argc-1; c++) 
	if (argv[c][0] != '-')
	    do_copy(argv[c],argv[argc-1]);
	
    bdos(14,Curdrive);			/* Reset the current drive */
}

/*  do_copy(source,destination)
    do_copy finds and copies files from the source path to the destination,
    figuring out which files to copy first to ensure an optimum placement
    on the destination media.
    
*/
    
do_copy(source,dest) 
char *source,*dest;
{
    char *files[MAXF],temp[100],*myalloc(),*tc,*scdir();
    int ff,x,t,user,count,curdrive;
    long sizes[MAXF],fsize(),freesp(),sprem;

    if (dest[1] == ':')
	curdrive = toupper(dest[0])-'A';
    else
	curdrive = Curdrive;
	
    				/* First, scan the specified directory */
    
    for (ff=x=0; x<MAXF; x++) {
	if (tc = scdir(source)) {
	    files[x] = myalloc(strlen(tc) + 1);
	    strcpy(files[x],tc);
	    sizes[x] = Fsize;
	    ff++;
	} else
	    break;
    }
	
    
    if (ff ) {			/* If any files were found, process them */
	sortsz(sizes,files,ff);	/* Sort into size order */
    
    				/* Sorted. Now copy to destination. */
	for (count=0,x = ff; x > 0; ) {	
	    sprem = freesp(curdrive);
	    for (t=ff-1; t>=0; t--) {
		if (sizes[t] != -1 && sprem > sizes[t]) {
		    copy(files[t],dest,sprem);
		    x--;
		    count = 0;
		    sizes[t] = -1;
		    break;
		}
	    }
	    if (t < 0) {
		if (++count > 1) {
		    printf("Could not mount sufficient space for:\n");
		    for (t=0; t<ff; t++)
			if (sizes[t] != -1) 
			    printf("%14s %6ldK\n",files[t],
				    (sizes[t]+1023L)/1024L);
		    exit(1);
		}
		for (sprem = 0L, t=0; t<ff; t++)
		    if (sizes[t] !=-1)
			sprem += sizes[t];
		
		printf("Please mount new disk on %s, %d files (%ldK) left to copy.\nPress Enter to continue."
			,dest,x,sprem/1024L);
		gets(temp);
	    }
	}
    }
}

/*  sortsz(sizes, files, number)
	    
    Sortsz sorts an array of longs into ascending sequence, and 'carries'
    the array files. Number is the total number of items to
    sort. sortsz uses a separate sort algorithm, since the built-in qsort
    does not have an external exchange facility, and thus cannot carry
    fields without using structures.
*/

sortsz(sizes,files,number)
char *files[];
long sizes[];
int number;
{
    register int i, j;
    int	gap, k;
    long temp1;
    char *temp2;

    for( gap=1; gap <= number; gap = 3*gap + 1 )
	;

    for( gap /= 3;  gap > 0  ; gap /= 3 ) {
	for( i = gap; i < number; i++ ) {
	    for( j = i-gap; j >= 0 ; j -= gap )	{
		if (sizes[j] > sizes[j + gap]) {
		    temp1=sizes[j]; sizes[j]=sizes[j+gap]; sizes[j+gap]=temp1;
		    temp2=files[j]; files[j]=files[j+gap]; files[j+gap]=temp2;
		}
	    }
	}
    }
}


char *myalloc(x)		/* Simply an error checking version of
    				   malloc()
    				 */
int x;
{
    char *malloc(),*s;
	
    if ((s=malloc(x))==NULL) {
	printf("Can't allocate %d bytes.\n",x);
	exit(1);
    }
    return(s);
}

/* copy(source, destination, bytes left)
    copy is the workhorse of the fill program, copying the source file to the
    destination. Note that a write error here is a fatal error, since the
    operating system just told us we had enough space for this file.
*/

copy(s,d,l)
char *s,*d;
long l;
{
    static char *buf;
    int fi,fo,len;
    long time,ftime();
    char t[128],temp[128];
    
    addpath(t,d,s);

    printf("%9ld bytes remaining. Copying %s to %s.\n",l,s,t);
    
    if ((fi=open(s,0))==-1) {
	fprintf(stderr,"Can't open %s for input!\n",s);
	exit(2);
    }
    time = ftime(0,fi);
    
    if (!Quiet) {
	if ((fo = open(t,0)) != -1) {
	    close(fo);
	    fprintf(stderr,"fill: %s already exists, overwrite? ",t);
	    gets(temp);
	    if (tolower(temp[0]) != 'y') {
		close(fi);
		exit(3);
	    }
	}
    }
    
    if ((fo = creat(t,0644)) == -1) {
	fprintf(stderr,"Can't create %s for write in copy!\n",t);
	exit(2);
    }
    close(fo);
    if ((fo = open(t,2)) == -1) {
	fprintf(stderr,"Can't re-open %s for write!\n",t);
	exit(2);
    }
    
    if (!buf)
	buf = myalloc(32640);
    
    len = read(fi,buf,32640);
    
    while (len > 0) {
	if (write(fo,buf,len) != len) {
	    fprintf(stderr,"Disk write error on file %s\n",t);
	    exit(3);
	}
	len = read(fi,buf,32640);
    }
    lseek(fo,0L,0);
    read(fo,buf,1);
    ftime(1,fo,time);
    close(fi);close(fo);
}

/* addpath(dest,src1,src2)

    addpath make a new filename from a destination path, plus a source
    path. For example, passed 'c:\temp' and 'd:/new/file/name/temp2',
    addpath will return in dest 'c:\temp\temp2'.
*/
    
addpath(dest,src1,src2)
char *dest,*src1,*src2;
{
    register char *t;
    char *strchr();
    
    strcpy(dest,src1);		/* start out in the right frame of mind */
    dest += strlen(dest)-1;	/* Point to the last char */
    if (strchr(":\\/",*dest++) == 0)	/* Add path sep. if not present */
	*dest++ = '\\';
    
    t = src2 + strlen(src2) - 1;	/* Point to last char */
    
    while ((t != src2) && (strchr(":\\/",*t) == 0))
	t--;				/* Point to last part of name */
    if (t != src2)
	t++;				/* Point past path sep. */
    
    strcpy(dest,t);			/* Add path to result */
}
    

/*  scdir(path)
    scdir returns the next matching file in the path, and returns the file's
    size in the global variable Fsize
*/

char *
scdir(pat)
char *pat;
{
    register char *cp, *np;
    char *strchr();
    static int code = SRCHFIR;
    static char nambuf[64], *tail;
    static struct {
	char reserve[21];
	char attr;
	long mtime;
	long fsize;
	char name[13];
    } srchbuf;

    if (code == SRCHFIR) {
	if (strchr(pat, '*') == 0 && strchr(pat, '?') == 0) {
	    bdos(SETDMA, &srchbuf);
	    if (dos(code, 0, 00, pat) == -1) 
		Fsize = 0;
	    else
		Fsize = srchbuf.fsize;
	    code = 0;
	    return(pat);
	}

	strncpy(nambuf, pat, 64);
	for (cp = tail = nambuf ; cp < &nambuf[64] && *cp ; ++cp)
	    if ((tail == nambuf && *cp == ':') || *cp == '/' || *cp == '\\')
		tail = cp+1;
    }

    bdos(SETDMA, &srchbuf);
    if (code == 0 || dos(code, 0, 00, pat) == -1) {
	code = SRCHFIR;
	return (char *)0;
    }
    code = SRCHNXT;

    np = tail;
    cp = srchbuf.name;
    while (*np++ = tolower(*cp++))
	;
    Fsize = srchbuf.fsize;
    return(nambuf);
}


/*  freesp(drive)
    
    freesp determines the free space remaining on the specified (zero based)
    drive. The result is in bytes.
*/

long freesp(dr)
int dr;
{
    long Kspace;
    char buffer[128];
    
    if (!Cpcdos4) {
    Kspace = Kspace;
#asm
    mov	dl,4[bp]	;get drive number
    inc	dl
    mov	ah,36h		;get free disk space
    int	21h		;go do it
    cmp	ax,-1		;is it an error?
    jz	freerr		;yes, get out
    xor	dx,dx		;make number of sectors per cluster long
    mul	bx		;get number of free sectors
    mul	cx		;get number of bytes free
freerr:
    mov	-2[bp],dx
    mov -4[bp],ax
#endasm
    } else {
    Kspace = Kspace;
#asm
	xor	dx,dx		;set drive mask to zero
	stc			;set carry
	mov	cx,4[bp]	;get drive code
	inc	cx
stdr:	rcl	dx,1		;rotate carry though register
	loop	stdr		;back for more
	mov	cl,37		;reset specific drive
	int	224		;do it
	push	es		;save current es
	mov	cl,52		;get current dma values
	int	224
	push	bx		;save them on stack
	push	es
	mov	dx,ss		;point to our segment
	mov	cl,51
	int	224
	lea 	dx,-132[bp]	;point to buffer area
	mov 	cl,26		;set dma buffer
	int 	224
	mov 	dl,4[bp]	;pick up drive code
	mov 	cl,46		;get free space
	int 	224
	pop	dx
	mov	cl,51		;point back to orginal segment
	int 	224
	pop	dx
	mov	cl,26		;point back to original offset
	int	224
	pop	es
#endasm
	Kspace = ((*(long *)buffer)&0xffffff) * 128L;
    }
    return(Kspace);
}

static char sbuf[100],*linebuf;

printf(fmt,args)
char *fmt;
unsigned args;
{
    int mputc();
    
    linebuf = sbuf;
    format(mputc,fmt,&args);
    mflush();
}

mputc(c)
char c;
{
    if (linebuf-sbuf > 99) 
	mflush();
    *linebuf++ = c;
}

mflush() {
    write(1,sbuf,linebuf-sbuf);
    linebuf=sbuf;
}
